//SRB2 Multi v2 - mod inspired by Sonic Multi (Sonic 1 hack made by MarkeyJester)
//Made by Nightwolf

local multicount = 0
//Multiplied object count (for object count limiting)

local isenabled = CV_RegisterVar({
	name = "srb2m_enable",
	defaultvalue = "Yes",
	flags = CV_NETVAR,
	PossibleValue = CV_YesNo,
	func = function() end
})

//If player's ring count is equal or higher, they will die
//Setting it to 0 disables the limit
local ringlimit = CV_RegisterVar({
	name = "srb2m_ringlimit",
	defaultvalue = tostring(250),
	flags = CV_NETVAR,
	PossibleValue = {MIN = 0, MAX = 9999},
	func = function() end
})

//Cooldown between object spawn and possible object multiplication
local multicooldown = CV_RegisterVar({
	name = "srb2m_cooldown",
	defaultvalue = tostring(TICRATE/5), //7
	flags = CV_NETVAR,
	PossibleValue = {MIN = 0, MAX = 2100},
	func = function() end
})

//Cooldown for picking up rings (on top of object cooldown; per player, not per object)
local ringcooldown = CV_RegisterVar({
	name = "srb2m_ringcooldown",
	defaultvalue = tostring(1),
	flags = CV_NETVAR,
	PossibleValue = {MIN = 0, MAX = 2100},
	func = function() end
})

//Multiplied object limit
//Setting it to 0 disables the limit
local multilimit = CV_RegisterVar({
	name = "srb2m_objectlimit",
	defaultvalue = tostring(1024),
	flags = CV_NETVAR,
	PossibleValue = {MIN = 0, MAX = 64000},
	func = function() end
})

local objspread = CV_RegisterVar({
	name = "srb2m_objspreadtype",
	defaultvalue = "RELEASE",
	flags = CV_NETVAR,
	PossibleValue = {DEV = 0, PRERELEASE = 1, RELEASE = 2},
	func = function() end
})

addHook("NetVars",function(net)
	multicount = net(multicount)
end)

//Enemy destroying cooldown (srb2m_cooldown)
addHook("MobjDamage",function(target, inflictor, source)
	if isenabled.value == 0 then return end
	if not source then 
		if inflictor then 
			source = inflictor
		else
			return
		end
	end
	if source.type != MT_PLAYER then return end
	if target.multicooldown then return true end
end)

//Monitor breaking cooldown
addHook("MobjMoveCollide",function(tmthing,thing)
	if isenabled.value == 0 then return end
	if thing.multicooldown and thing.flags & MF_MONITOR then return false end
end, MT_PLAYER)

//Ring collecting cooldown (both object cooldown and ring cooldown are checked here)
addHook("TouchSpecial", function(special, toucher)
	if isenabled.value == 0 then return end
	if special.type != MT_RING and special.type != MT_FLINGRING then return end
	if special.multicooldown or toucher.mringcdown then return true end
end)

//Offensive comment

//Make destroyed monitors disappear after 1 second and disappearing rings lower object count properly
addHook("MobjFuse",function(mo)
	if mo.multimonitor then
		mo.state = S_NULL
		return true
	elseif mo.type == MT_RING or mo.type == MT_FLINGRING then
		if mo.multicount and multicount > 0 then 
			multicount = $1 - 1
			mo.state = S_NULL
			return true
		end
	end
end)

//The most important part: object multiplication and stuff
addHook("MobjDeath",function(target, inflictor, source)
	if isenabled.value == 0 then return end
	if target.flags & MF_BOSS != 0 or G_IsSpecialStage() or target.type == MT_EGGTRAP then return end
	if target.multicount and multicount > 0 then multicount = $1 - 1 end
	if target.multimonitor then return end
	if not source then 
		if inflictor then 
			source = inflictor
		else
			return
		end
	end
	if source.type != MT_PLAYER then return end
	local obj1
	local obj2
	if target.type == MT_RING then
		local obj0 = P_SpawnMobj(target.x,target.y,target.z,MT_RING)
		obj0.multicooldown = multicooldown.value
		if multicount >= multilimit.value-1 and multilimit.value != 0 then return end
		obj1 = P_SpawnMobj(target.x,target.y,target.z+target.height,MT_FLINGRING)
		obj1.multicooldown = multicooldown.value
		obj1.multicount = true
		multicount = $1 + 1
	else
		if multicount >= multilimit.value-1 and multilimit.value != 0 then return end
		if target.type == MT_GFZFISH then
			obj1 = P_SpawnMobj(target.x,target.y,target.z+target.height,target.type)
			obj2 = P_SpawnMobj(target.x,target.y,target.z+target.height,target.type)
		elseif P_RandomChance(FRACUNIT/2)
			obj1 = P_SpawnMobj(target.x-target.radius,target.y,target.z+target.height,target.type)
			obj2 = P_SpawnMobj(target.x+target.radius,target.y,target.z+target.height,target.type)
		else
			obj1 = P_SpawnMobj(target.x,target.y-target.radius,target.z+target.height,target.type)
			obj2 = P_SpawnMobj(target.x,target.y+target.radius,target.z+target.height,target.type)
		end
		obj1.multicooldown = multicooldown.value
		obj2.multicooldown = multicooldown.value
		obj1.multicount = true
		obj2.multicount = true
		multicount = $1 + 2
	end
	
	local minx, miny, rchance
	if objspread.value == 0 then minx = 4; miny = 4; rchance = 0 //Development version object X/Y spread emulation (some gifs were made during that phase)
	elseif objspread.value == 1 then minx = 4; miny = 4; rchance = FRACUNIT/2 //Prerelease version object X/Y spread emulation
	else minx = 0; miny = 0; rchance = FRACUNIT/2 end //Current behaviour
	
	if obj1 then
		obj1.momx = P_RandomRange(minx,64)*FRACUNIT/8
		if P_RandomChance(rchance) then obj1.momx = -$1 end
		obj1.momy = P_RandomRange(miny,64)*FRACUNIT/8
		if P_RandomChance(rchance) then obj1.momy = -$1 end
		if obj1.flags & MF_NOGRAVITY == 0 then obj1.momz = P_RandomRange(2,32)*FRACUNIT/8 end
		if obj1.type == MT_FLINGRING then obj1.fuse = 4*TICRATE; source.mringcdown = ringcooldown.value end
	end
	if obj2 then
		obj2.momx = P_RandomRange(minx,64)*FRACUNIT/8
		if P_RandomChance(rchance) then obj2.momx = -$1 end
		obj2.momy = P_RandomRange(miny,64)*FRACUNIT/8
		if P_RandomChance(rchance) then obj2.momy = -$1 end
		if obj2.flags & MF_NOGRAVITY == 0 then obj2.momz = P_RandomRange(2,32)*FRACUNIT/8 end
		if obj2.type == MT_FLINGRING then obj2.fuse = 4*TICRATE end
	end
	if target.flags & MF_MONITOR != 0 then
		target.multimonitor = true
		target.fuse = TICRATE
	end
	//print("MC: "..tostring(multicount))
end)

//Reset multiplied object count on map change
addHook("MapChange",function(mapnum)
	multicount = 0
end)

//Object cooldown stuff
addHook("MobjThinker",function(mo)
	if isenabled.value == 0 then return end
	if mo.multicooldown then 
		mo.multicooldown = $1 - 1
		if mo.flags & MF_MONITOR then mo.flags = $1 & ~MF_SOLID end
	elseif mo.multicooldown == 0 and mo.info.flags & MF_SOLID then
		mo.flags = $1 | MF_SOLID
	end
	if mo.multiremove then mo.state = S_NULL end //Remove flickies completely
end)

//Flag flickies for later removal
addHook("MobjSpawn",function(mo)
	if isenabled.value == 0 then return end
	if mo.type >= MT_FLICKY_01 and mo.type <= MT_SECRETFLICKY_02_CENTER then mo.multiremove = true; return true end
end)

//Player behaviour
addHook("PlayerThink",function(p)
	if isenabled.value == 0 then return end
	if not p.mo then
		return
	elseif p.mo.mringcdown then 
		p.mo.mringcdown = $1 - 1
	end
	if ringlimit.value == 0 then return end
	if p.rings >= ringlimit.value then P_DamageMobj(p.mo,nil,nil,1,DMG_INSTAKILL) end 
end)